export declare type BaseDB<Schema> = Collections<Schema> & {
    transaction: TransactionFun<Schema>;
    query: QueryFun;
};
export declare type SchemaConstraint = {
    [key: string]: CollectionConstraint;
};
export declare type Column = DBRaw;
declare type TransactionFun<Schema> = (trx: (db: BaseDB<Schema>) => Promise<void>, rollback?: () => Promise<void>) => Promise<void>;
declare type Collections<Schema> = {
    [P in keyof Schema]: Schema[P] extends CollectionConstraint ? Collection<Schema[P]> : never;
};
export declare type QueryResult<T, Keys extends keyof T, CF extends string> = ([Keys] extends [never] ? T : Pick<T, Keys>) & {
    [P in CF]: string | undefined;
};
declare type DBValue = DBValueBase | DBValueBase[];
declare type DBValueBase = string | number | boolean | Date | undefined;
declare type Keys<T> = Extract<keyof T, string>;
export declare type CollectionConstraint = {
    id: string;
    [key: string]: DBValue;
};
declare type WhereOr<T extends {
    id: string;
}> = Where<T> | Where<T>[];
declare type NumOperators<T = number | DBQuery> = {
    ne?: T;
    gt?: T;
    gte?: T;
    lt?: T;
    lte?: T;
    between?: [T, T];
    notBetween?: [T, T];
    in?: number[];
    notIn?: number[];
};
declare type DateOperators<T = Date | DBQuery> = {
    ne?: T;
    gt?: T;
    gte?: T;
    lt?: T;
    lte?: T;
    between?: [T, T];
    notBetween?: [T, T];
};
declare type StrOperators<T = string | DBQuery> = {
    in?: string[];
    ne?: T;
    gt?: T;
    gte?: T;
    lt?: T;
    lte?: T;
    like?: T;
    notLike?: T;
    iLike?: T;
    notILike?: T;
    regexp?: T;
    notRegexp?: T;
    iRegexp?: T;
    notIRegexp?: T;
};
declare type ArrOperators<T = number | DBQuery> = {
    contains?: T;
    ne?: T;
    gt?: T;
    gte?: T;
    lt?: T;
    lte?: T;
    overlap?: T;
    contained?: T;
};
declare type WhereItem<T> = T extends number ? NumOperators | number : (T extends string ? StrOperators | T : (T extends Date ? DateOperators : (T extends Array<infer V> ? ArrOperators<V[]> | V[] : never)));
declare type Where<T> = {
    [P in keyof T]?: T[P] | WhereItem<T[P]> | DBQuery;
} | DBQuery;
declare type Other<T, Fields extends Keys<T>, CustomFields extends string> = {
    select?: ReadonlyArray<Fields>;
    selectCustom?: {
        [P in CustomFields]: DBQuery;
    };
    order?: {
        desc?: Keys<T>;
        asc?: Keys<T>;
    };
    limit?: number;
    offset?: number;
};
export declare class DBEntityNotFound extends Error {
    entityName: string;
    cond: string;
    constructor(entityName: string, cond: string);
}
export declare class DBQueryError extends Error {
    query: string;
    values: DBValue[];
    error: string;
    constructor(query: string, values: DBValue[], error: string);
}
export declare function sql(strs: TemplateStringsArray, ...inserts: QueryValue[]): DBQuery;
export declare function joinQueries(queries: DBQuery[], separator?: DBQuery): DBQuery;
declare class Collection<T extends CollectionConstraint> {
    collectionName: string;
    private query;
    private loader;
    name: Column;
    fields: {
        [P in keyof T]: Column;
    };
    constructor(collectionName: string, query: QueryFun, prevCollection: Collection<T> | undefined);
    private loadById;
    findById<K extends Keys<T> = never>(id: T['id'], other?: {
        select?: K[];
    }): Promise<QueryResult<T, K, never>>;
    findOne<K extends Keys<T> = never, CF extends string = never>(where: WhereOr<T>, other?: Other<T, K, CF>): Promise<QueryResult<T, K, CF>>;
    findAll<K extends Keys<T> = never, CF extends string = ''>(where: WhereOr<T>, other?: Other<T, K, CF>): Promise<QueryResult<T, K, CF>[]>;
    findByIdOrNull<K extends Keys<T> = never>(id: T['id'], other?: {
        select?: K[];
    }): Promise<QueryResult<T, K, never> | undefined>;
    findOneOrNull<K extends Keys<T> = never, CF extends string = never>(where: WhereOr<T>, other?: Other<T, K, CF>): Promise<QueryResult<T, K, CF> | undefined>;
    update(id: T['id'], data: {
        [P in keyof T]?: T[P] | DBQuery | (T[P] extends number ? NumberUpdate : never);
    }): Promise<void>;
    remove(id: T['id']): Promise<void>;
    create(data: {
        [P in keyof T]: T[P] | DBQuery | (P extends 'id' ? ('auto' | T[P]) : T[P]);
    }, params?: {
        noErrorIfConflict?: Keys<T> | DBQuery | true;
    }): Promise<void>;
    genId(): T["id"];
}
declare type NumberUpdate = {
    increment: number;
} | {
    decrement: number;
};
declare type Pool = {
    connect: () => Promise<PoolClient>;
};
declare type PoolClient = {
    release: () => void;
    query: (q: string, values?: unknown[]) => Promise<{
        rows: unknown[];
        command: string;
    }>;
};
declare type QueryFun = <T>(query: DBQuery) => Promise<T[]>;
export declare function createDB<Schema extends SchemaConstraint>(pool: Pool): Promise<BaseDB<Schema>>;
declare type QueryValue = DBValue | DBRaw | DBQuery | DBQueries;
declare class DBRaw {
    readonly raw: string;
    constructor(raw: string);
}
declare class DBQuery {
    private readonly parts;
    private readonly values;
    constructor(parts: ReadonlyArray<string>, values: ReadonlyArray<QueryValue>);
}
declare class DBQueries {
    readonly queries: ReadonlyArray<DBQuery>;
    readonly separator: DBQuery | undefined;
    constructor(queries: ReadonlyArray<DBQuery>, separator: DBQuery | undefined);
}
export interface Migration {
    up: string;
    name: string;
}
export declare function readMigrationsFromDir(dir: string): Promise<Migration[]>;
export declare function migrateUp(db: BaseDB<unknown>, migrations: Migration[], logger: {
    info: (...args: unknown[]) => void;
}): Promise<void>;
export {};
//# sourceMappingURL=PostgresqlDriver.d.ts.map